import time
import logging
import logging.handlers
import socket
import threading
import yaml
import traceback
from urllib.parse import parse_qs
import queue 
from zeep.client import Client, CachingClient, Settings
from zeep.wsse.username import UsernameToken
import zeep.helpers
import datetime

# Initialize global variables and locks
lock = threading.Lock()
shutdown_event = threading.Event()
queries_info = {}
active_threads = 0
zmapi = None
monitors = None

alerted = []
alerted_lock = threading.Lock()

network_lock = threading.Lock()

# Initialize two thread-safe queues
queue1 = queue.Queue()
queue2 = queue.Queue()

settings = Settings()
settings.strict = False
settings.xml_huge_tree = True

# # WSDL File
url = "https://www.onvif.org/ver10/events/wsdl/event.wsdl"

import os

# Define the path to the config file
config_file_path = '/etc/zm/onvifserv.yaml'

# Check if the config file exists
if not os.path.exists(config_file_path):
    print(f"Error: The configuration file {config_file_path} was not found.")
    exit(1)  # Exit if the config file does not exist

# Load configuration from YAML file
with open(config_file_path, 'r') as config_file:
    config = yaml.safe_load(config_file)

if config is None:
    print("Error: Configuration is empty or incorrectly formatted.")
    exit(1)  # Exit if the configuration is not loaded properly

def setup_logging():
    if 'logpath' in config:
        log_filename = config['logpath']
    else:
        print("Logpath not found in config. Using default 'app.log'.")
        log_filename = 'app.log'
    
    log_file_size_limit = config.get('log_file_size_limit', 1048576)  # Use a default value if not specified
    
    # Ensure the log directory exists
    os.makedirs(os.path.dirname(log_filename), exist_ok=True)
    
    # Set up a logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    
    # Create handlers
    c_handler = logging.StreamHandler()
    f_handler = logging.handlers.RotatingFileHandler(
        log_filename, maxBytes=log_file_size_limit, backupCount=5)
    
    # Create formatters and add it to handlers
    c_format = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    f_format = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    c_handler.setFormatter(c_format)
    f_handler.setFormatter(f_format)
    
    # Add handlers to the logger
    logger.addHandler(c_handler)
    logger.addHandler(f_handler)
    
    return logger


logger = setup_logging()

def send_command(command):
    try:
        with network_lock:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                # Connect to the server with a timeout
                s.settimeout(5)  # 5 seconds timeout
                s.connect((config.get("zmtriggerhost", "localhost"), config.get("zmtriggerport", 6802)))
                # Send the command followed by a newline, and encode to bytes
                s.sendall(f"{command}\n".encode())
    except socket.timeout:
        logger.error(f"Timeout occurred when sending command: {command}")
    except Exception as e:
        logger.error(f"An error occurred when sending command: {command}. Error: {str(e)}")

def alert_on(monitor, cause, score=1, info='',showtext=''):
    if monitor in config.get("monitors",{}).keys():
        id = config.get("monitors",{})[monitor]['id']
        command = str(id)+"|on|"+str(score)+"|"+cause+"|"+info+"|"+showtext
        send_command(command)
    else:
        logger.error("Attempted to switch on a monitor that doesn't exist! Monitor name:"+monitor)

def alert_off(monitor, cause, score=1, info='',showtext=''):
    if monitor in config.get("monitors",{}).keys():
        id = config.get("monitors",{})[monitor]['id']
        command = str(id)+"|cancel||||"
        send_command(command)
    else:
        logger.error("Attempted to switch off a monitor that doesn't exist! Monitor name:"+monitor)

def observeMonitor(name, ip, user, password, port):
    global alerted_lock
    timeout = datetime.timedelta(seconds=config.get('polltimeout',300))
    subtimeout = datetime.timedelta(seconds=config.get('subtimeout',3600))

    # # *** Events Service ***
    xaddr = "http://"+ip+"/onvif/events_service"
    zeep_client_events = CachingClient(wsdl=url, wsse=UsernameToken(user, password, use_digest=True), settings=settings)
    ws_client_events = zeep_client_events.create_service("{http://www.onvif.org/ver10/events/wsdl}EventBinding", xaddr)

    zeep_client_pp = CachingClient(wsdl=url, wsse=UsernameToken(user, password, use_digest=True), settings=settings)
    ws_client_pp = zeep_client_pp.create_service("{http://www.onvif.org/ver10/events/wsdl}PullPointSubscriptionBinding", xaddr)

    termtime = 0
    while not shutdown_event.is_set():
        if termtime < time.time():
            res = ws_client_events.CreatePullPointSubscription(InitialTerminationTime=subtimeout)
            subManagerAddress=res['SubscriptionReference']['Address']['_value_1']
            zeep_client_sub_mngr = CachingClient(wsdl=url, wsse=UsernameToken(user, password, use_digest=True), settings=settings)
            ws_client_sub_mngr = zeep_client_sub_mngr.create_service("{http://www.onvif.org/ver10/events/wsdl}PullPointSubscriptionBinding", subManagerAddress)
            logger.info(f"Subscribed to events on:"+ip)
            termtime = time.time()+config.get('subtimeout',3600)
        try:
            res=ws_client_sub_mngr.PullMessages(MessageLimit=1, Timeout=timeout)
            if res and 'NotificationMessage' in dir(res) and len(res['NotificationMessage']) > 0:
                for entry in res['NotificationMessage']:
                    if 'Message' in dir(entry) and '_value_1' in dir(entry['Message']) and len(entry['Message']['_value_1']) > 0:
                        for el in entry['Message']['_value_1']:
                            if 'Data' in str(el) and len(el) > 0:
                                si = el[0]
                                if 'Name' in si.keys() and 'Value' in si.keys() and 'IsMotion' in si.get('Name'):
                                    if 'true' in si.get('Value'):
                                        logger.info(f"Motion start received on:"+ip)
                                        with lock:
                                            query_data = {
                                                "timestamp": int(time.time())
                                            }
                                            queries_info[ip] = query_data
                                        with alerted_lock:
                                            if name not in alerted:
                                                queue1.put(name)
                                                logger.info(f"Motion start submit on:"+ip)
                                    else:
                                        logger.info(f"Motion stop received on:"+ip)
                                        with lock:
                                            if ip in queries_info.keys() and queries_info[ip]['timestamp']+config.get("minalert",30) < int(time.time()):
                                                logger.info(f"Motion stop submit on:"+ip)
                                                queue2.put(name)
                                            else:
                                                logger.info(f"Too fast motion stop on:"+ip)
        except Exception as e:
            logger.info(traceback.format_exc())
    ws_client_sub_mngr.Unsubscribe()



def check_ongoing():
    while not shutdown_event.is_set():
        with alerted_lock:
            with lock:
                for alerted_item in alerted:
                    if alerted_item in config.get('monitors', dict()).keys() and \
                        config.get('monitors', dict())[alerted_item]['ip'] in queries_info.keys() and \
                        int(time.time()) - queries_info[config.get('monitors', dict())[alerted_item]['ip']]['timestamp'] > config.get('monitors', dict())[alerted_item]['expiry']:
                        queue2.put(alerted_item)
        shutdown_event.wait(10)  # Wait for 10 seconds or until the shutdown event is set

def notifier():
    while not shutdown_event.is_set():

        try:
            # Process items from queue2
            item2 = queue2.get_nowait()
            logger.info(f"Cancelling alert on: {item2}")
            alert_off(item1, "Cam Event")
            queue2.task_done()
            with alerted_lock:
                alerted.remove(item2)
        except queue.Empty:
            pass
        except Exception:
            logger.error("Fail in cancel alert queue processing!")

        try:
            # Process items from queue1
            item1 = queue1.get_nowait()
            logger.info(f"Alerting on: {item1}")
            alert_on(item1, "Cam Event")
            queue1.task_done()
            with alerted_lock:
                alerted.append(item1)
        except queue.Empty:
            pass
        except Exception:
            logger.error("Fail in alert queue processing!")
        # Wait for an item to be added to one of the queues
        if queue1.empty() and queue2.empty():
            time.sleep(0.1)  # Polling interval

if __name__ == "__main__":
    alive_thread = threading.Thread(target=check_ongoing)
    alive_thread.start()

    notifier_thread = threading.Thread(target=notifier)
    notifier_thread.start()

    monitor_threads = []  # Keep track of monitor threads

    for monitor_name, monitor_info in config.get('monitors', {}).items():
        t = threading.Thread(target=observeMonitor, args=(monitor_name, monitor_info['ip'], monitor_info['username'], monitor_info['password'], monitor_info['port']))
        t.start()
        monitor_threads.append(t)

    try:
        for t in monitor_threads:
            t.join()
    except KeyboardInterrupt:
        logger.info("Shutting down...")
        shutdown_event.set()
        alive_thread.join()
        notifier_thread.join()

